import pyewjn.util

from typing import Callable
import numpy as np

SMALL_X_BOUNDARY = 1e3


def get_zeta_p_integrand(
	eps: Callable[[float], complex]
) -> Callable[[float, float], complex]:
	"""Gets the integrand function zeta_p_integrand(u, y).

	Returns zeta_p_integrand(u, y), a complex valued function of two momenta in units of vacuum wavelength.

	:param eps:
	:return:
	"""

	def zeta_p_integrand(y: float, u: float) -> complex:
		"""
		Here y and u are in units of vacuum wavelength, coming from Ford-Weber / from the EWJN noise expressions.
		:param u:
		:param y:
		:return:
		"""
		u2 = u**2
		y2 = y**2
		k2 = u2 + y2
		k = np.sqrt(k2)
		eps_value = eps(k)
		term_1 = y2 / (eps_value - k2)
		term_2 = u2 / eps_value
		return (term_1 + term_2) / k2

	return zeta_p_integrand


def get_zeta_p_function(eps: Callable[[float], complex]):
	def integrand1_small_x(x: float, u: float) -> complex:
		u2 = u**2
		x2 = x**2
		eps_value = eps(u * np.sqrt(1 + x**2))
		return (x2 / (eps_value - 1 - x2)) / ((1 + x2) * u2)

	def integrand1_big_x(x: float, u: float) -> complex:
		u2 = u**2
		x2 = x**2
		eps_value = eps(u * x)
		return 1 / ((eps_value - x2) * u2)

	# 1 / (eps(u * sqrt(1 + x^2)))   * 1 / (1 + x^2)

	def integrand2_small_x(x: float, u: float) -> complex:
		x2 = x**2
		eps_value = eps(u * np.sqrt(1 + x**2))
		return 1 / (eps_value * (1 + x2))

	def integrand2_big_x(x: float, u: float) -> complex:
		x2 = x**2
		eps_value = eps(u * x)
		return 1 / (eps_value * x2)

	def zeta_p(u: float) -> complex:
		# zeta_p_integrand = get_zeta_p_integrand(eps)

		i1_small = pyewjn.util.complex_quad(
			lambda x: integrand1_small_x(x, u), 0, SMALL_X_BOUNDARY, epsabs=1e-12
		)
		i1_big = pyewjn.util.complex_quad(
			lambda x: integrand1_big_x(x, u), SMALL_X_BOUNDARY, np.inf, epsabs=1e-12
		)
		i2_small = pyewjn.util.complex_quad(
			lambda x: integrand2_small_x(x, u), 0, SMALL_X_BOUNDARY, epsabs=1e-12
		)
		i2_big = pyewjn.util.complex_quad(
			lambda x: integrand2_big_x(x, u), SMALL_X_BOUNDARY, np.inf, epsabs=1e-12
		)

		integral = sum(term[0] for term in [i1_small, i2_small, i1_big, i2_big])

		return integral * 2j * u

	return zeta_p



from typing import Callable

import numpy as np
import scipy.integrate

import pyewjn.noise.im_ref


def get_chi_zz_e(eps: Callable[[float], complex]) -> Callable[[float], float]:
	im_ref_p = pyewjn.noise.im_ref.get_im_ref_p(eps)

	def chi_zz_e(z: float) -> float:
		def integrand(y: float) -> float:
			return (y**2) * im_ref_p(y / z) * np.exp(-2 * y)

		integral = scipy.integrate.quad(
			integrand, 0, np.inf, epsabs=1e-10, epsrel=1e-10
		)
		return integral[0] / (z**3)

	return chi_zz_e



from typing import Callable

import numpy as np

import pyewjn.noise.zeta


def get_im_ref_p(eps: Callable[[float], complex]) -> Callable[[float], float]:
	zeta_p = pyewjn.noise.zeta.get_zeta_p_function(eps)

	def im_ref_p(u: float) -> float:
		zeta_p_val = zeta_p(u)
		return np.imag((np.pi * 1j * u - zeta_p_val) / (np.pi * 1j * u + zeta_p_val))

	return im_ref_p



import numpy as np
from scipy.integrate import quad, quadrature


def complex_quad(func, a, b, **kwargs):
	def real_func(x):
		return np.real(func(x))

	def imag_func(x):
		return np.imag(func(x))

	real_integral = quad(real_func, a, b, **kwargs)
	imag_integral = quad(imag_func, a, b, **kwargs)

	return (
		real_integral[0] + 1j * imag_integral[0],
		real_integral[1:],
		imag_integral[1:],
	)


def complex_quadrature(func, a, b, **kwargs):
	def real_func(x):
		return np.real(func(x))

	def imag_func(x):
		return np.imag(func(x))

	real_integral = quadrature(real_func, a, b, **kwargs)
	imag_integral = quadrature(imag_func, a, b, **kwargs)

	return (
		real_integral[0] + 1j * imag_integral[0],
		real_integral[1:],
		imag_integral[1:],
	)



import numpy as np
import pyewjn.dielectric.sigma_nam
import pyewjn.dielectric.low_k_nam

from pyewjn.baskets import CalculationParams, CalculationConstants

from typing import Tuple

FIXED_LARGE_MOMENTUM = 1e8


class DedimensionalisedParameters(object):
	def __init__(
		self,
		omega: float,
		sigma_n: float,
		tau: float,
		v_f: float,
		temp: float,
		critical_temp: float,
		c_light: float,
	):
		gap = 0
		if temp < critical_temp:
			# else, problems will happen
			gap = 3.06 * np.sqrt(critical_temp * (critical_temp - temp))
		self.xi = omega / gap
		self.nu = 1 / (tau * gap)
		self.t = temp / gap
		self.a = omega * v_f / (c_light * gap)
		self.b = sigma_n / omega


class NamDielectricCoefficients(object):
	def __init__(self, a: float, b: float, c: float, d: float):
		self.a = a
		self.b = b
		self.c = c
		self.d = d
		self.u_l = np.real((-self.c + 1j * self.d) / (-self.a + 1j * self.b))

	def eps(self, u_c: float):
		def piecewise_eps(u: float):
			# todo add check for u_c vs u_l
			if u < self.u_l:
				return -self.a + 1j * self.b
			elif self.u_l < u < u_c:
				return 1 + (-self.c + 1j * self.d) / u
			else:
				return 1

		return piecewise_eps


def get_dedimensionalised_parameters(
	omega: float,
	sigma_n: float,
	tau: float,
	v_f: float,
	temp: float,
	critical_temp: float,
	c_light: float,
) -> DedimensionalisedParameters:
	return DedimensionalisedParameters(
		omega, sigma_n, tau, v_f, temp, critical_temp, c_light
	)


def get_small_momentum_coefficients(
	dedim_params: DedimensionalisedParameters,
) -> Tuple[float, float]:
	prefactor = 4j * np.pi * dedim_params.b
	s = pyewjn.dielectric.low_k_nam.sigma_nam_alk(
		dedim_params.xi, 0, dedim_params.nu, dedim_params.t
	)
	conductivity = prefactor * s
	return -np.real(conductivity), np.imag(conductivity)


def get_big_momentum_coefficients(
	dedim_params: DedimensionalisedParameters,
) -> Tuple[float, float]:
	prefactor = 4j * np.pi * dedim_params.b * FIXED_LARGE_MOMENTUM / dedim_params.a
	s = pyewjn.dielectric.sigma_nam.sigma_nam(
		dedim_params.xi, FIXED_LARGE_MOMENTUM, dedim_params.nu, dedim_params.t
	)
	conductivity = prefactor * s
	return -np.real(conductivity), np.imag(conductivity)


def get_nam_dielectric_coefficients(
	omega: float,
	sigma_n: float,
	tau: float,
	v_f: float,
	temp: float,
	crit_temp: float,
	c_light: float,
) -> NamDielectricCoefficients:
	"""Gets a NamDielectricCoefficients object, using SI unit parameters

	:param omega: frequency
	:param sigma_n: normal state conductivity
	:param tau: tau in Hz
	:param v_f: Fermi velocity, in m/s
	:param temp: temperature in Hz
	:param crit_temp: critical temperature, in Hz
	:param c_light: speed of light, meters per second
	:return:
	"""

	dedim = get_dedimensionalised_parameters(
		omega, sigma_n, tau, v_f, temp, crit_temp, c_light
	)
	a, b = get_small_momentum_coefficients(dedim)
	c, d = get_big_momentum_coefficients(dedim)

	return NamDielectricCoefficients(a, b, c, d)


def get_nam_dielectric(
	u_c: float,
	params: CalculationParams,
	constants: CalculationConstants = CalculationConstants(),
):
	if params.omega is None:
		raise ValueError("omega expected to not be None")
	if params.v_f is None:
		raise ValueError("v_f expected to not be None")
	if params.omega_p is None:
		raise ValueError("omega_p expected to not be None")
	if params.tau is None:
		raise ValueError("tau expected to not be None")
	if params.t_rel is None:
		raise ValueError("relative temp expected to not be None")
	if params.t_c is None:
		raise ValueError("critical temp expected to not be None")
	sigma_n = params.omega_p**2 * params.tau / (4 * np.pi)
	coeffs = get_nam_dielectric_coefficients(
		params.omega,
		sigma_n,
		params.tau,
		params.v_f,
		params.t_rel * params.t_c,
		params.t_c,
		constants.c_light,
	)
	return coeffs.eps(u_c)



from pyewjn.dielectric.nam_dielectric_coefficient_approximator import get_nam_dielectric
from pyewjn.dielectric.lindhard_dielectric import get_lindhard_dielectric

__all__ = ["get_nam_dielectric", "get_lindhard_dielectric"]



class CalculationConstants(object):
	"""Holds physical constants, in SI units"""

	def __init__(
		self,
		epsilon_0=8.854e-12,
		h_bar=1.0546e-34,
		c_light=3e8,
		electron_mass=9.10938356e-31,
	):
		"""Initialises constants in SI units, with sensible defaults

		:param epsilon_0:
		:param h_bar:
		:param c_light:
		:param electron_mass:
		"""

		self.epsilon_0 = epsilon_0
		self.h_bar = h_bar
		self.c_light = c_light
		self.electron_mass = electron_mass


class CalculationParams(object):
	"""Holds the parameters describing a calculation, in SI units."""

	def __init__(
		self,
		omega: float = None,
		omega_p: float = None,
		tau: float = None,
		v_f: float = None,
		t_rel: float = None,
		t_c: float = None,
		dipole_moment: float = None,
	):
		"""Creates parameter object, SI units

		:param omega:
		:param omega_p:
		:param tau:
		:param v_f:
		:param t_rel:
		:param t_c:
		:param dipole_moment:
		"""

		self.omega = omega
		self.omega_p = omega_p
		self.tau = tau
		self.v_f = v_f
		self.t_rel = t_rel
		self.t_c = t_c
		self.dipole_moment = dipole_moment



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyewjn', 'pyewjn.dielectric', 'pyewjn.noise', 'pyewjn.util']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.1,<2.0.0', 'scipy>=1.8,<1.9']

setup_kwargs = {
    'name': 'pyewjn',
    'version': '0.2.0',
    'description': '',
    'long_description': None,
    'author': 'Deepak',
    'author_email': 'dmallubhotla+github@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import numpy as np
from numpy.lib.scimath import sqrt as csqrt

import pyewjn.util


def g(w, wp):
	return ((wp * (w + wp)) + 1) / (csqrt(wp**2 - 1) * csqrt((w + wp) ** 2 - 1))


def f(k, e, v):
	return ((4 / 3) * 1 / (e - 1j * v)) + (4 / 15) * (1 / ((e - 1j * v) ** 3)) * k**2


def i1(w, wp, k, v):
	gv = g(w, wp)
	e1 = csqrt((w + wp) ** 2 - 1)
	e2 = csqrt(wp**2 - 1)

	f_upper = f(k, np.real(e1 - e2), np.imag(e1 + e2) + 2 * v) * (gv + 1)
	f_lower = f(k, np.real(-e1 - e2), np.imag(e1 + e2) + 2 * v) * (gv - 1)

	return f_upper + f_lower


def i2(w, wp, k, v):
	gv = g(w, wp)
	e1 = csqrt((w + wp) ** 2 - 1)
	e2 = csqrt(wp**2 - 1)

	f_upper = f(k, np.real(e1 - e2), np.imag(e1 + e2) + 2 * v) * (gv + 1)
	f_lower = f(k, np.real(e1 + e2), np.imag(e1 + e2) + 2 * v) * (gv - 1)

	return f_upper + f_lower


def a(w, k, v, t):
	return pyewjn.util.complex_quad(
		lambda wp: np.tanh((w + wp) / (2 * t)) * (i1(w, wp, k, v)), 1 - w, 1
	)[0]


def b_int(wp, w, k, v, t):
	return (np.tanh((w + wp) / (2 * t)) * i1(w, wp, k, v)) - (
		np.tanh(wp / (2 * t)) * i2(w, wp, k, v)
	)


def b(w, k, v, t, b_max=np.inf):
	return pyewjn.util.complex_quad(lambda wp: b_int(wp, w, k, v, t), 1, b_max)[0]


def sigma_nam_alk(w, k, v, t):
	return -1j * (3 / 4) * (v / w) * (-a(w, k, v, t) + b(w, k, v, t))



import numpy as np
from pyewjn.baskets import CalculationConstants, CalculationParams

LINDHARD_SERIES_THRESHOLD = 1e4


class LindhardDielectric(object):
	def __init__(
		self,
		params: CalculationParams,
		constants: CalculationConstants = CalculationConstants(),
		thres=LINDHARD_SERIES_THRESHOLD,
	):
		if params.omega is None:
			raise ValueError("omega expected to not be None")
		if params.v_f is None:
			raise ValueError("v_f expected to not be None")
		if params.omega_p is None:
			raise ValueError("omega_p expected to not be None")
		if params.tau is None:
			raise ValueError("tau expected to not be None")

		self.series_threshold = thres
		self.omega = params.omega
		self.v_f = params.v_f
		self.omega_p = params.omega_p
		self.tau = params.tau
		self.c_light = constants.c_light

		self.s = 1 / (self.tau * self.omega)
		self.prefactor = 3 * (self.omega_p**2) / (self.omega**2)

	def get_eps(self):
		def eps_lindhard(u_inverse_wavelength: float) -> complex:
			"""the lindhard dielectric function

			:param u_inverse_wavelength: u is in units of the reciprocal vacuum wavelength (omega / c_light)
			:return: returns the value of epsilon, dimensionless
			"""

			# converts u from inverse vacuum wavelength to inverse mean free path
			u = u_inverse_wavelength * self.v_f / self.c_light

			if u < LINDHARD_SERIES_THRESHOLD * self.v_f / self.omega:
				return eps_series(u)
			else:
				return eps_full_lindhard(u)

		def eps_series(u: float) -> complex:
			rel_value = (1j / (3 * (self.s - 1j))) + (
				u**2 * ((-9j + 5 * self.s) / (45 * (-1j + self.s) ** 3))
			)

			return 1 + (self.prefactor * rel_value)

		def eps_full_lindhard(u: float) -> complex:

			log_value = np.log((1 - u + (self.s * 1j)) / (1 + u + (self.s * 1j)))

			rel_value = (1 + ((1 + (self.s * 1j)) / (2 * u)) * log_value) / (
				1 + ((self.s * 1j) / (2 * u)) * log_value
			)

			return 1 + (self.prefactor / (u**2)) * rel_value

		return eps_lindhard


def get_lindhard_dielectric(
	params: CalculationParams, constants: CalculationConstants = CalculationConstants()
):
	return LindhardDielectric(params, constants).get_eps()
